/*
	Copyright 2013-2014, Marten de Vries

	Licensed under the Apache License, Version 2.0 (the "License");
	you may not use this file except in compliance with the License.
	You may obtain a copy of the License at

	http://www.apache.org/licenses/LICENSE-2.0

	Unless required by applicable law or agreed to in writing, software
	distributed under the License is distributed on an "AS IS" BASIS,
	WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
	See the License for the specific language governing permissions and
	limitations under the License.
*/

/*global gate */
"use strict";

exports.objects = {};
exports.functions = [];

var RECURSION_LIMIT = 10;

exports.log = function () {
  if (arguments.length === 1) {
    console.log("DEBUG: " + JSON.stringify(arguments[0]));
  } else {
    console.log("DEBUG: " + JSON.stringify(toArray(arguments)));
  }
};

//based on https://gist.github.com/nolanlawson/10340255.
function createBlob(parts, properties) {
  properties = properties || {};
  try {
    return new global.Blob(parts, properties);
  } catch (e) {
    if (e.name !== "TypeError") { //coverage: ignore
      throw e;
    }
    var BlobBuilder = ( //coverage: ignore
      global.BlobBuilder ||
      global.MSBlobBuilder ||
      global.MozBlobBuilder ||
      global.WebKitBlobBuilder
    );
    var builder = new BlobBuilder(); //coverage: ignore
    for (var i = 0; i < parts.length; i += 1) { //coverage: ignore
      builder.append(parts[i]); //coverage: ignore
    }
    return builder.getBlob(properties.type); //coverage: ignore
  }
}

//based on: http://stackoverflow.com/a/16245768
exports.b64toBlob = function(b64Data) {
  var sliceSize = 1024;

  function charCodeFromCharacter(c) {
    return c.charCodeAt(0);
  }

  var byteCharacters = global.atob(b64Data);
  var byteArrays = [];

  for (var offset = 0; offset < byteCharacters.length; offset += sliceSize) {
    var slice = byteCharacters.slice(offset, offset + sliceSize);
    var byteNumbers = Array.prototype.map.call(slice, charCodeFromCharacter);
    var byteArray = new global.Uint8Array(byteNumbers);

    byteArrays.push(byteArray.buffer);
  }
  return createBlob(byteArrays);
};

var callbackCache = {};
exports.createCallback = function (id) {
  if (typeof callbackCache[id] === "undefined") {
    callbackCache[id] = function() {
      var func = modifyArg.bind(null, null, 0);
      //converting arguments with toArray() seems required in
      //older versions of webkit.
      asyncMap(func, toArray(arguments), function (result) {
        result.length = result._count;
        var callbackArgs = toArray(result);
        gate.callbackCalled(id, JSON.stringify(callbackArgs));
      });
    };
  }
  return callbackCache[id];
};

function asyncMap(func, data, callback) {
  var result = {_count: 0};
  var count = 0;

  function createMapCallback(key) {
    return function (item) {
      result[key] = item;
      result._count += 1;
      if (count === result._count) {
        callback(result);
      }
    };
  }
  var key;
  for (key in data) {
    count += 1;
  }
  for (key in data) {
    var mapCallback = createMapCallback(key);
    try {
      func(data[key], mapCallback);
    } catch (e) {
      //Just pass in undefined instead of crashing.
      //
      //(Context specific for Python-PouchDB):
      //Most likely it's just an XHR object that
      //generates an InvalidStateError while
      //serializing some property of it that no user
      //would ever want to access...
      mapCallback(undefined); //coverage: ignore
    }
  }
  if (!count) {
    callback(result);
  }
}

function modifyArg(owner, recursionDepth, arg, callback) {
  if (recursionDepth > RECURSION_LIMIT) {
    callback("recursion limit reached");
  } else if (arg instanceof global.Blob) {
    blobToBase64(arg, function (base64) {
      callback({
        type: "_js_returned_blob",
        blobType: arg.type,
        data: base64
      });
    });
  } else if (typeof arg === "object" && arg !== null && arg.constructor !== Date) {
    //recursion
    var func = modifyArg.bind(null, arg, recursionDepth + 1);
    asyncMap(func, arg, function (newArg) {
      if (arg instanceof Array) {
        newArg.length = newArg._count;
        newArg = toArray(newArg);
      } else if (arg instanceof Error) {
        newArg.message = arg.toString();
      }
      delete newArg._count;
      callback(newArg);
    });
  } else if (typeof arg === "function") {
    var id = exports.functions.push(arg.bind(owner)) -1;
    callback({
      type: "_js_returned_function",
      functionId: id
    });
  } else {
    callback(arg);
  }
}

function blobToBase64(blob, cb) {
  var reader = new global.FileReader();
  reader.onload = function() {
    var dataUrl = reader.result;
    var base64 = dataUrl.split(',')[1];
    cb(base64);
  };
  //a native blob
  reader.readAsDataURL(blob);
}

function toArray(arrayLike) {
  return Array.prototype.slice.call(arrayLike, 0);
}
