# coding=UTF-8
#
#	Copyright 2013-2014, Marten de Vries
#
#	Licensed under the Apache License, Version 2.0 (the "License");
#	you may not use this file except in compliance with the License.
#	You may obtain a copy of the License at
#
#	http://www.apache.org/licenses/LICENSE-2.0
#
#	Unless required by applicable law or agreed to in writing, software
#	distributed under the License is distributed on an "AS IS" BASIS,
#	WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#	See the License for the specific language governing permissions and
#	limitations under the License.

import contextlib
import pouchdb

@contextlib.contextmanager
def suppress(*e):
	try:
		yield
	except e:
		pass

class PromiseStub(object):
	def __init__(self, *args, **kwargs):
		super(PromiseStub, self).__init__(*args, **kwargs)

		self._successHandlers = set()
		self._failureHandlers = set()

	def then(self, h):
		self._successHandlers.add(h)

	def catch(self, h):
		self._failureHandlers.add(h)

	def success(self, resp):
		for h in self._successHandlers:
			h(resp)

	def failure(self, err):
		"""Assumes ``err`` to be sys.exc_info()'s output."""

		if not self._failureHandlers:
			raise err[1], None, err[2]
		for h in self._failureHandlers:
			h(err[1])

class JSON(object):
	def __init__(self, data):
		self.data = data

class Blob(object):
	def __init__(self, data):
		self.data = data

class JSError(Exception):
	def __init__(self, messages, *args, **kwargs):
		message = "\n".join(messages)

		super(JSError, self).__init__(message, *args, **kwargs)

class AttrAccessDict(dict):
	def __getattr__(self, key):
		try:
			return super(AttrAccessDict, self).__getattr__(key)
		except AttributeError:
			try:
				return self[key]
			except KeyError, e:
				raise AttributeError(e)

def getPromiseResult(ctx, promise):
	returned = {}
	def success(resp):
		returned["resp"] = resp
	def fail(err):
		returned["err"] = err

	promise.then(success, fail)
	ctx.waitUntilCalled([success, fail])

	if returned.get("err"):
		raise pouchdb.PouchDBError(returned["err"])
	return returned["resp"]

@contextlib.contextmanager
def gettingCallbackResult(ctx):
	returning = {}
	def callback(err=None, resp=None):
		returning["err"] = err
		returning["resp"] = resp
	info = {"callback": callback}
	yield info
	ctx.waitUntilCalled(callback)
	if returning.get("err"):
		raise pouchdb.PouchDBError(returning["err"])
	info["resp"] = returning["resp"]

def filterNone(args):
	return [a for a in args if a is not None]

def jsonAllowed(value):
	if isinstance(value, basestring):
		return JSON(value)
	return value

def aliasFor(method):
	methodName = method.__name__
	def alias(self, *args, **kwargs):
		return getattr(self, methodName)(*args, **kwargs)
	alias.__doc__ = "Alias for :meth:`%s`." % methodName
	return alias

@contextlib.contextmanager
def jsToPouchError():
	try:
		yield
	except JSError, e:
		raise pouchdb.PouchDBError(e)

@contextlib.contextmanager
def jsTypeErrorToAttributeError():
	try:
		yield
	except JSError, e:
		if "TypeError" in str(e):
			raise AttributeError(e)
		raise #pragma: no cover
